/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/* Written by Elizabeth S. Odekirk */

package FP;

import hamlet.*;
import java.awt.*;
import java.util.*;
import java.applet.Applet;
import java.io.*;
import java.awt.Toolkit;

public class Eqn extends Panel {

  String outStr;
  StreamTokenizer st;
  String sval;
  TextField eqnField = new TextField("2 + 3 = 5                             ");
  TextField statusField = new TextField("0 roundings                     ");
  Panel eqnPanel     = new Panel();
  Panel statusPanel  = new Panel();
  final int PLUS = '+';
  final int MINUS = '-';
  FP parent;
  int mantVal = 1, minExpVal = -1, maxExpVal = 1;
  RdsFrame rdsFrame = new RdsFrame();
  StringBuffer execString = new StringBuffer();
  
  double maxx = 0;
  double minx = 0;
  int numRoundings = 0;
  Node root;
  int offsetStrLen = 0;
  String fonts[];

  public Eqn(FP parentIn) {
    /* The expression line */
    eqnPanel.setLayout(new BorderLayout());
    eqnPanel.add("West", new Label("Expression Input "));
    eqnPanel.add("Center", eqnField);   
    /* The status line */
    statusField.setEditable(false);
    statusPanel.setLayout(new BorderLayout());
    statusPanel.add("West", new Label("Expression Status"));
    statusPanel.add("Center", statusField);
    /* The whole component */
    setLayout(new BorderLayout());
    add("North", eqnPanel);
    add("South", statusPanel);
    parent = parentIn;
  } /* end init */  

  /* What to do when they hit return in the text field,
   * or a key because they are going to start editting */
  public boolean handleEvent( Event e ) {
    int eqidx;
    String exptxt;

    if (e.target == eqnField) {
      if (e.id == Event.ACTION_EVENT)  {
	parseIt();
	return true;
      }
      if (e.id == Event.KEY_PRESS) {
	int start, end;
	exptxt = eqnField.getText();
	eqidx = exptxt.indexOf('=');
	if (eqidx >= 0) {
	  start = eqnField.getSelectionStart();
	  end = eqnField.getSelectionEnd();
	  exptxt = exptxt.substring(0, eqidx);
	  eqnField.setText(exptxt);
	  eqnField.select(start, end);
	  return false;
	}
	return false;
      } /* end KeyPress */
    } /* end eqnField */
    return false;
  } /* end handleEvent() */


  public void reset (String s) {
    statusField.setText(s);
    eqnField.setText("");
  }

  public void parseIt() {
    String streq;
    String strval;
    int token = 0;
    int eqidx = 0;
    String typedEqn;
    
    /* This is the new node of our tree */
    root = new Node ('e');
    /* Set our message to be blank */
    rdsFrame.rds.setText("");
    typedEqn = eqnField.getText();
    /* Do this to get rid of the answer.  When people are typing
     * in a new equation, we don't want the old equation sitting there.
     * You always want what shows up on the screen to be true. */
    eqidx = typedEqn.indexOf('=');
    if (eqidx >= 0)
      typedEqn = typedEqn.substring(0, eqidx);
    /*System.out.println("The initial string is: " + typedEqn);*/
    st = new StreamTokenizer(new StringBufferInputStream(typedEqn));
    st.ordinaryChar('/');  
    st.ordinaryChar('-');  
    st.ordinaryChar('e');
    mantVal = parent.sharedInfo.mantSize;
    maxExpVal = parent.sharedInfo.maxExp;
    minExpVal = parent.sharedInfo.minExp;
    minx = Math.pow(10, minExpVal);
    maxx = Math.pow(10, maxExpVal+1) * (1 - Math.pow(10, -(mantVal)));
    Val.parent = this;
    Node.parent = this;
    statusField.setText("");
    numRoundings = 0;
    rdsFrame.rds.setText("");
    try {
      root = exp();
      offsetStrLen = (root.toString()).length();
      offsetStrLen += 5;
      token = st.nextToken();
      if (token != StreamTokenizer.TT_EOF) 
	throw new SyntaxException("extra tokens");
      eval(root);
    } catch (IOException ex) {}
      /* This means that either they typed in something gramatically wrong
       * or one of the numbers they entered cannot be represented with the
       * number system parameters */
      catch (SyntaxException e) {
	eqidx = typedEqn.indexOf('=');
	if (eqidx >= 0) {
	  typedEqn = eqnField.getText();
	  eqidx = typedEqn.indexOf('=');
	  if (eqidx >= 0)
	    typedEqn = typedEqn.substring(0, eqidx);
	}
	eqnField.setText(typedEqn);
	statusField.setText("Syntax Error - " + e.getMessage());
	return;
      }
      catch (ComputationException e) {
	statusField.setText("Computation Error - " + e.getMessage());
	return;
      }
      
    
    if (root.left != null)
      root = root.left;
    rdsFrame.rds.appendText(root.toString());
    /*System.out.println("");*/
      
    /*System.out.println("Value is: " + val);*/
    outStr = typedEqn.trim();
    outStr = outStr.concat(" = ");
    outStr = outStr.concat(root.val.toString());
    eqnField.setText(outStr);
    if (numRoundings == 1) 
      statusField.setText("1 rounding");
    else
      statusField.setText(numRoundings + " roundings");
  }
  
  /* This is where we begin calculating the equation. */

  /* Expression.  + and - */
  Node exp() throws SyntaxException, ComputationException {
    Node retNode;
    Node val;
    int token = 0;

    val = term();
    while (true) {
      try {
	token = st.nextToken();
      } catch (IOException ex) {}
      if (token == '+' ) {
	retNode = new Node('+');
	retNode.left  = val;
	retNode.right = term();
	/* This needs to change - won't support
	 * several +s in a row. */
	val = retNode;
      }
      else if (token == '-' ) {
	retNode = new Node('-');
	retNode.left  = val;
	retNode.right = term();
	val = retNode;
      }
      else {
	st.pushBack();
	return val;
      }
    }
  } /* end exp */

  /* Executing * and / */
  Node term() throws SyntaxException, ComputationException {
    Node retNode;
    Node val;
    int token = 0;

    val = foo();
    while (true) {
      try {
	token = st.nextToken();
      } catch (IOException ex) {}
      if (token == '*') {
	retNode = new Node('*');
	retNode.left  = val;
	retNode.right = foo();
	val = retNode;
      }
      else if (token == '/') {
	retNode = new Node('/');
	retNode.left  = val;
	retNode.right = foo();
	val = retNode;
      }
      else {
	st.pushBack();
	return val;
      }
    }
  } /* end term */

  /* Executing unary -  and + */
  Node foo() throws SyntaxException, ComputationException {
    Node    retNode;  
    Node    val;
    boolean gotUnaryMinus = false;
    int     token = 0;

    try {
      token = st.nextToken();
    } catch (IOException ex) {}
    if (token == '-') 
      gotUnaryMinus = true;
    else if (token == '+')
      ;
    else 
      st.pushBack();
    val = pt();
    if (gotUnaryMinus) {
      retNode = new Node('u');
      retNode.left = val;
      return retNode;
    }
    else
      return val;
  } /* end foo */

  /* Executing ^ */
  Node pt()  throws SyntaxException, ComputationException {
    Node retNode;
    Node val;
    int token = 0;

    val = f();
    try {
      token = st.nextToken();
    } catch (IOException ex) {}
    if (token == '^') {
      retNode = new Node('^');
      retNode.left  = val;
      retNode.right = foo();
      return (retNode);
    }
    else 
      st.pushBack();
    return val;
  } /* end pt */

  /* Executing ( and ) and number */
  Node f() throws SyntaxException, ComputationException {
    Node retNode;
    Node val = new Node('e');
    int token = 0;
    
    try {
      token = st.nextToken();
    } catch (IOException ex) {}
    if (token == '(') {
      val = exp();
      retNode = new Node ('(');
      retNode.left = val;
      val = retNode;
      try {
	token = st.nextToken();
      } catch (IOException ex) {}
      if (token != ')') 
	throw new SyntaxException("no )");
    }
    else if (st.ttype == StreamTokenizer.TT_NUMBER) {
      String newfoo;
      int exp = 0;
      double prenum, tryexp;
      boolean expIsNeg = false;
      prenum = st.nval;
      /* Looking to see if there is scientific notation */
      try {
	token = st.nextToken();
      } catch (IOException ex) {}
      if (st.ttype == 'e') {
	/*System.out.println("This has an exp");*/
	try {
	  token = st.nextToken();
	} catch (IOException ex) {}
	/* Unary plus or minus in scientific notation exponent */
	if (token == '-' || token == '+') {
	  if (token == '-')
	    expIsNeg = true;
	  try {
	    token = st.nextToken();
	  } catch (IOException ex) {}
	}
	if (token == StreamTokenizer.TT_EOF)
	  throw new SyntaxException ("no exponent in scientific notation");
	if (st.ttype != StreamTokenizer.TT_NUMBER) 
	  throw new SyntaxException ("exponent isn't a number");
	tryexp =  st.nval;
	exp = (int) st.nval;
	if ((double) exp != st.nval)
	  throw new SyntaxException ("e exponent isn't " +
				     "an integer");
	if (expIsNeg)
	  exp = -exp;
	/*System.out.println("Scientific exponent is " + exp);*/
	prenum = prenum * Math.pow(10, exp);
	retNode = new Node (prenum);
      }
      else {
	retNode = new Node (prenum);
	st.pushBack();
      }
      return retNode;
    }
    else {
      throw new SyntaxException("number expected");
    }
    return val;
  } /* end f */

  
  /* Return a string with blanks which is 
   * offsetStrLen - curLen long.  It's used
   * to put in a tab stop at offsetStrLen */
  String blankStrToOffset(int curLen) {
    StringBuffer str = new StringBuffer();
    int i;
    for (i = curLen; i < offsetStrLen; i++)
      str.append(" ");
    return (str.toString());
  }

  /* This is used to print out the sub-node which
   * is currently being executed and 
   * which goes to the side of the full tree listing */
  String subnodeStr(int curStrLen, Node n) {
    StringBuffer str = new StringBuffer();
    str.append(blankStrToOffset(curStrLen));
    str.append(n.toString());
    return (str.toString());
  }

  /* Evaluate the nodes.  
   *   If the node is just a number, return it. 
   *   If the node is an operator, return the number
   *     which the operator creates. 
   *   
   *   The node for () only has a left hand side, and
   *     that side holds */
  Node eval(Node n) throws ComputationException, SyntaxException {
    Node   retVal = new Node ('e');
    String str    = "";
    int    curStrLen = 0, i;
    String outStr;

    if (n.type == Node.OPER_TYPE) {
      n.left = eval(n.left);
      if (n.right != null) 
	n.right = eval(n.right);
      /*System.out.println("");*/
      outStr = root.toString();
      curStrLen = outStr.length();
      rdsFrame.rds.appendText(root.toString());
      /*rdsFrame.rds.appendText("\n");*/
      switch (n.oper) {
        case ('+'):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  (n.left).val.add(n.right.val);
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  str = retVal.val.check();
	  break;
	case ('-'):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  (n.left).val.subtr(n.right.val);
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  str += retVal.val.check();
	  break;
	case ('*'):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  (n.left).val.mult(n.right.val);
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  str += retVal.val.check();
	  break;
	case ('/'):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  (n.left).val.div(n.right.val);
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  str += retVal.val.check();
	  break;
	case ('u'):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  (n.left).val.negate();
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  str += retVal.val.check();
	  break;
	case ('^'):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  (n.left).val.expon(n.right.val);
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  str += retVal.val.check();
	  break;
	case ('('):
	  rdsFrame.rds.appendText(subnodeStr(curStrLen, n));
	  rdsFrame.rds.appendText(" = " + n.left.val.toString() );
	  n = n.left;
	  retVal = n;
	  break;
      } /* end of switch */
    } 
    else if (n.type == Node.NUM_TYPE) {
      retVal = n;
      str = retVal.val.checkInput();
    }
    rdsFrame.rds.appendText("\n");
    if (!str.equalsIgnoreCase("")) {
      for (i = 0; i < (offsetStrLen+3); i++)
	rdsFrame.rds.appendText(" ");
      rdsFrame.rds.appendText(str);
      str = new String ("");
      numRoundings++;
    }
    return retVal;
  }
}

/* This is window which holds the 
 * execution listing */
class RdsFrame extends Frame {
  Panel    area = new Panel();
  TextArea rds = new TextArea(200, 200); 
  TextArea lds = new TextArea(200, 200);
  Button   closeButton = new Button ("Close");

  public Dimension preferredSize() {
    return new Dimension(220,220);
  }


  public RdsFrame (){
    Font newFont;

    newFont = new Font("Courier", 
		       Font.PLAIN, 12);
    rds.setFont(newFont);
    Panel closePanel = new Panel();
    closePanel.setLayout(new FlowLayout(FlowLayout.LEFT));
    closePanel.add(closeButton);

    area.setLayout(new GridLayout(1, 2, 2, 2));
    area.add(rds);
    area.add(lds);

    add ("North", closePanel);
    add ("Center", rds);
    /*add ("Center", area);*/
    setTitle("Execution Window");
    pack();
    /*show();*/
  }

  public boolean action(Event e, Object arg) {
    if (e.target.equals(closeButton) ) {
      hide();
      return true;
    }
    return false;
  }
}

/* The two types of exceptions which I throw */
class SyntaxException extends Exception {

  public SyntaxException(String s) {
    super(s);
  }
}

class ComputationException extends Exception {

  public ComputationException(String s) {
    super(s);
  }
}


/* A node in the parse tree. */
class Node {

  Node  left;
  Node  right;
  int   type;
  char  oper;
  Val   val;
  
  static Eqn parent;

  static final int OPER_TYPE = 1;
  static final int NUM_TYPE  = 2;

  public Node (char operIn) {
    constNode();
    oper = operIn;
    type = OPER_TYPE;
  }

  public Node (double valIn) {
    constNode();
    val  = new Val(valIn);
    type = NUM_TYPE;
  }
    

  void constNode() {
    left  = null;
    right = null;
    val   = null;
    oper  = 'x';
  }

  public String toString() {
    StringBuffer str = new StringBuffer("");

    if (type == OPER_TYPE && oper == '(') {
      str.append(" (");
      /*parent.rdsFrame.rds.appendText(" (");*/
      /*System.out.print(" (");*/
      str.append(left.toString());
      str.append(" )");
      /*parent.rdsFrame.rds.appendText(" )");*/
      /*System.out.print(" )");*/
    }
    else {
      if (left != null) 
	str.append(left.toString());
      if (type == OPER_TYPE) {
	if (oper == 'u') {
	  str.append(" -");
	  /*parent.rdsFrame.rds.appendText(" -");*/
	  /*System.out.print(" -");*/
	}
	else {
	  str.append(" " + oper);
	  /*parent.rdsFrame.rds.appendText(" " + oper);*/
	  /*System.out.print(" " + oper);*/
	}
      }
      else if (type == NUM_TYPE) {
	str.append(" " + val.toString());
	/*parent.rdsFrame.rds.appendText(" " + val.toString());*/
	/*System.out.print(" " + val.toString());*/
      }
      if (right != null)
	str.append(right.toString());
    }
    return(str.toString());
  }
} /* end class Node */

class Val {
  double value;

  static Eqn parent;

  public Val(double valueIn) {
    value = valueIn;
  }
  
  public String toString() {
    String str;
    str = new String("" + value);
    return str;
  }

  public void add(Val valTwo) {
    value += valTwo.value;
  }

  public void subtr(Val valTwo) {
    value -= valTwo.value;
  }

  public void mult(Val valTwo) {
    value *= valTwo.value;
  }

  public void div(Val valTwo) {
    value /= valTwo.value;
  }

  public void negate() {
    value = -value;
  }

  public void expon(Val valTwo) {
    value = Math.pow(value, valTwo.value);
  }

  public String check() throws ComputationException, SyntaxException {
    return(insideCheck(false));
  }

  public String checkInput() throws ComputationException, SyntaxException {
    return (insideCheck(true));
  }

  /* Check to see that the number is in our number system */
  String insideCheck(boolean isInput) 
                         throws ComputationException, SyntaxException {
    int retval = 0;
    String retString = new String ("");
    double preVal;
    double tempVal = 0;
    double valLog  = 0,
           rndLog  = 0;
    double inVal;
    double maxx, minx;
    int mantVal;
    boolean rounded = false;

    preVal = value;
    maxx = parent.maxx;
    minx = parent.minx;
    mantVal = parent.mantVal;

    /* Number too high */
    if (Math.abs(value) > maxx) {
      /*System.out.print("high: " + value);*/
      if (isInput)
	throw new SyntaxException(value + " too large");
      else
	throw new ComputationException(value + " too large");
    }

    /* Number too low */
    else  if (Math.abs(value) < minx && 
	Math.abs(value) != 0) {
      /*System.out.print("low: " + value);*/
      
      /*System.out.println(" : " + value);*/
      if (isInput)
	throw new SyntaxException(value + " too small");
      else
	throw new ComputationException(value + " too small");
    }

    /* Check to see if, because of mantissa size,  
     * we need to round it */
    else {
      double fooval;
      valLog  = Math.floor(Math.log(Math.abs(value))/Math.log(10.0));
      rndLog  = mantVal - (1 + valLog);
      tempVal = value * Math.pow(10, rndLog);
      /*System.out.print("rnd: " + value);*/
      /*System.out.print(" : " + tempVal);*/
      fooval = tempVal;
      tempVal = Math.round(tempVal);
/*      System.out.println("Original Value is " + fooval +
			 " new value is " + tempVal +
			 " difference is " + (fooval - tempVal));*/
      if (fooval != tempVal) {
 	if (Math.abs(fooval - tempVal) > Math.pow(10, -12)) {
	  rounded = true;
	}
      }
      /*System.out.print(" : " + tempVal);*/
      value   = tempVal * Math.pow(10, -rndLog);
      /*System.out.println(" : " + value);*/
    }

    if (rounded) {
      if (isInput)
	throw new SyntaxException(preVal + " not represented");
      else
	retString = new String (preVal + " -> " + value + "\n");
    }
    return (retString);
  } /* end check */

}
