/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/* 
<applet 
    CODEBASE=".."
    CODE="hamlet.AppButton.class" NAME="FP.FP"
    width=160 height=25 align=absbottom>
    <param name=LABEL value="Floating-Point Numbers">
    A Java applet should appear here
</applet>
*/

/* Written by Elizabeth S. Odekirk */

package FP;

import hamlet.*;
import java.awt.*;
import java.util.*;
import java.applet.Applet;

public class FP extends AppFrame implements Runnable {
	

    // -------------------------------------------------------------
    // Instance variables required for all appShells
    //     But, they don't need to go by these names.	
    // -------------------------------------------------------------
    
  String 	appletName;	      /* this applets name */
  String 	delimiters;	      /* place to store the delimiters */
  Thread 	appThread     = null; /* the application thread	     */

  final public String version  = "v1.0";
    
  Panel    enumeration, input, 
           top, bottom, statusLinePanel,
           enumLabels, enumAndLabel, enum,
	   numSystemPanel;
  ScrollableNumbers 
           mantSize, minExp, maxExp;
  TextArea mantEnum, floatEnum;
  MyCanvas plot = new MyCanvas();
  Button   incrPlot, decrPlot;
  Label    statusLine;
  MenuBar  menubar;
  Menu	   windowMenu;
  MenuItem executionItem;
  Eqn	   eqn;

  int      counter = 0;
  boolean  firstTime = true;
  boolean  doPlotInCalc = true;
  int      enumType = 0; /* 0:   summary enum
		            1:   full enum */
  boolean  justUp = true;
  FPSharedInfo sharedInfo = new FPSharedInfo();

  /* This is used to keep the original xmin and xmax for the given
   * mantSize, minExp and maxExp, so that the min and max can be
   * reset at will */
  double   stateXmin, stateXmax;

  final int fullEnum = 0;

  /* default values */
  final int    mantSizeDef = 1;
  final int    minExpDef   = -1;
  final int    maxExpDef   = 1;

  public FP() 
  {

    /*resize(900,900);*/

    setLayout(new BorderLayout());  

    setUpScreen();
    setUpHelp();

    appThread = new Thread(this);
    appThread.start();
    requestFocus();

    
  }

    //**************************************************************
    //  Start() runs each time the applet runs
    //**************************************************************/

  public void start()
  {
    appThread = new Thread(this);
    appThread.start();
    requestFocus();
  }

  public synchronized void stop()
  {
    if (appThread != null)
      appThread.stop();
    appThread     = null;	// Kill the application thread 	
  }


    //**************************************************************
    //  appStop() runs each time that the app is told to do this
    //            function via the controller applets.  (ie,
    //            Hamlet messages)
    //**************************************************************/

  public void appStop() {
    /*System.out.println(appletName+" appStop called");*/
    if (appThread != null)
      appThread.stop();
    appThread = null;          // Kill the application thread
  }

  //**************************************************************
  //  appStart() runs each time that the app is told to do this
  //             function via the controller applets. (ie, 
  //             Hamlet messages)
  //**************************************************************/
    
  public void appStart() {
    /* Insert your start code here */
    appThread = null;
    if (appThread == null) 
      appThread = new Thread(this);
    appThread.start();
    doPlotInCalc = true;
    /*appReset();*/
    requestFocus();
    /*    showStatus("Start the Applet");*/
  }


  //**************************************************************
  //  appReset() runs each time that the app is told to do this
  //             function via the controller applets.
  //**************************************************************

  public void appReset() {
    /* Insert your reset code here */
    /*      System.out.println(appletName+" appReset called");*/
      mantSize.setIntText(mantSizeDef);
      minExp.setIntText(minExpDef);
      maxExp.setIntText(maxExpDef);
      doPlotInCalc = true;
      appThread = new Thread(this);
      /*      showStatus("Reset the Applet");*/
  }


  /* This takes a message from a send button (this is
   * in the hamlet package) on the web page, and
   * parses it.  
   *
   * handles messages of the format:
   *   "recalc mantissa <int> minexp <int> maxexp <int>"
   *   "replot start <double> stop <double>"
   *   "eqn <equation string>"
   *
   * Anything else will give an error on stdout.
   */
  protected void handleMessage (String msg) {
    StringTokenizer st = new StringTokenizer(msg);
    String curToken;
    int mantsize = -1;
    int minexp = -1, maxexp = -1;
    int gotArgs = 0;
    int i;
    double minx = 0, maxx = 100;

    try {
      while (st.hasMoreTokens()) {
	curToken = st.nextToken();
	if (curToken.equalsIgnoreCase("recalc")) {
	  /* need all three arguments */
	  gotArgs = 0;
	  for (i=0; i<3; i++) { 
	    curToken = st.nextToken();
	    if (curToken.equalsIgnoreCase("Mantissa")) {
	      mantsize = (Integer.valueOf(st.nextToken())).intValue();
	      gotArgs = gotArgs + 1;
	    }
	    if (curToken.equalsIgnoreCase("Minexp")) {
	      minexp = (Integer.valueOf(st.nextToken())).intValue();
	      gotArgs = gotArgs + 2;
	    }
	    if (curToken.equalsIgnoreCase("Maxexp")) {
	      maxexp = (Integer.valueOf(st.nextToken())).intValue();
	      gotArgs = gotArgs + 3;
	    }
	  }
	  if (gotArgs != 6) {
	    System.out.println("Error in message - wrong parameters");
	    System.out.println("Msg: " + msg);
	    return;
	  }
	  mantSize.setIntText(mantsize);
	  minExp.setIntText(minexp);
	  maxExp.setIntText(maxexp);
	  appStart();
	}  /* End of recalc section */
      else if (curToken.equalsIgnoreCase("replot")) {
	for (i = 0; i < 2; i++) {
	  curToken = st.nextToken();
	  if (curToken.equalsIgnoreCase("Start")) {
	    minx = (Double.valueOf(st.nextToken())).doubleValue();
	    plot.setXmin(minx);
	  }
	  if (curToken.equalsIgnoreCase("Stop")) {
	    maxx = (Double.valueOf(st.nextToken())).doubleValue();
	    plot.setXmax(maxx);
	  }
        } /* end of for */
	plot.repaint();
      
      }/* End of replot */

      else if (curToken.equalsIgnoreCase("eqn")) {
	msg = msg.substring(3);
	msg = msg.trim();
	eqn.eqnField.setText(msg);
	eqn.parseIt();
	break;			// break out of the while loop
      }
      else {
        System.out.println("Unknown command.  Use Recalc, Replot, or Eqn.");
        System.out.println("Msg: " + msg);
	return;
      }

      } /* End of While */

      } /* End of try */
      catch (NoSuchElementException e) {
        System.out.println("Error in message - format error");
        System.out.println("Msg: " + msg + "; Error token: " + e.getMessage());
        return;
      }
      catch (NumberFormatException e) {
        System.out.println("Error in message - format error in number:");
        System.out.println("Msg: " + msg);
        System.out.println("Error token: " + e.getMessage());
        return;
      }
    /*System.out.println("Got to the end of handleMessage");
    System.out.println("Msg: " + msg);
    System.out.println("Mantissa Size: " + mantsize + 
                       " Minexp: " + minexp + 
                       " Maxexp: " + maxexp ); */
    
  }

 /* This adds a label with the given string and
   * the given constraints. */
  void addLabel(String name, GridBagLayout gridbag,
		GridBagConstraints c, Panel panel) {
    Label label = new Label(name);
    gridbag.setConstraints(label, c);
    panel.add(label);
  }
 

  void setUpHelp() {
    helpText = 
      "This applet graphs and lists the floating point numbers\n"	+ 
      "which can be formed using a mantissa of a given size and\n"	+
      "a minimum and maximum exponent.  It is based on the\n"		+
      "discussion in Chapter 2 of \"Introduction to Scientific\n"	+
      "Programming\" by Joseph L. Zachary.\n\n"				+ 
      //								//
      "File Menu\n"							+
      "   + Reset:  Returns to the default plot region\n"		+
      "   + Refresh: Refreshes the applet\n"				+
      "   + Hide:  Removes the applet from the display\n"		+
      "   + Quit:  Terminates the applet\n\n"				+
      //								//
      "Window Menu\n"							+
      "   + Execution: This window traces the execution of the\n"	+
      "         expression input in the expression field (see below).\n"+
      "         On the left side is the expression after \n"		+
      "         each operation is completed.  On the right is the\n"	+
      "         current operation.  If the result of that operation\n"  +
      "         is not one of the numbers represented, but does not\n"	+
      "         go over the maximum value or under the minimum value, \n"+
      "         then it is rounded to the nearest represented value.\n"	+
      "         This is shown by <badNumber> -> <goodNumber> on the\n"	+
      "         right side.\n\n"					+
      //								//
      "Help Menu\n"							+
      "   + With Applet:  Displays this help dialog\n\n"		+
      //								//
      "Fields \n"							+ 
      "   + Mantissa: choose the mantissa size\n"			+
      "   + Min Exponent: choose the minimum exponent\n"		+
      "   + Max Exponent: choose the maximum exponent\n"		+
      "\n"								+
      //								//
      "Expression\n"							+
      "   + Input: You can enter an expression here, and see if it\n"	+
      "        could complete with the given constraints on\n"		+
      "        mantissa size, maximum exponent and minimum\n"		+
      "        exponent.  You can use +, -, *, /, and ^ \n"		+
      "        (exponentiation), as well as ( and ).  You can \n"	+
      "        also type in a single number to see if it is in\n"	+
      "        the number system.\n"					+
      "   + Status: If the expression completes successfully, this will\n"+
      "        say \"Successfully Completed\".  There are two\n"	+
      "        kinds of errors, syntax errors and computation errors.\n"+
      "        A syntax error means that your input is incorrect;\n"	+
      "        either one of the numbers you put in is not in the\n"	+
      "        current number system, or you typed in an\n"		+
      "        incorrect format, such as two slashes (/) in a \n"	+
      "        row.  A computation error means that your input was\n"	+
      "        correct, but one of the interim results in solving\n"	+
      "        the expression, or the final result, could not be \n"	+
      "        represented.\n\n"					+
      //								//
      "Plot \n"								+
      "   + <<: decreases the maximum of the plot\n"			+ 
      "         by an order of magnitude\n"				+
      "   + >>: increases the maximum of the plot\n"			+
      "         by an order of magnitude\n"				+
      "   + dragging the mouse: \tcreate a zoom-in rectangle\n\n"	+ 
      //								//
      "Displayed Information\n"						+
      "  + Possible Mantissas: Lists the mantissas which can\n"		+
      "    be made with the current mantissa size\n"			+
      "  + Possible Floating Point Numbers: Lists  the numbers\n"	+
      "    which can be produced using the current mantissa size\n"	+
      "    and the range of exponents between minexp and maxexp\n"	+
      "  + Plot: Plot of the floating point list of numbers\n\n"	+
      //								//
      "Examples\n"							+
      "  + With a mantissa size of 1, your mantissa can\n"		+
      "    represent any integer from 1 to 9 (the first digit cannot\n" +
      "    be 0).  With a mantissa size of 2 your mantissa can\n"	+
      "    be more accurate, incrementing by 10ths.\n"			+ 
      "  + With a mantissa size of 1 and a maxexp of 1, you can\n"	+
      "    represent number up to 90.  If you raise the maxexp\n"	+
      "    to 2, you can represent numbers up to 900."			+
      //								//
      "\n";
  }

 void setUpScreen() {
   /*      showStatus("Setting up the screen");*/

    setTitle("Floating Point Numbers");

    setLayout(new BorderLayout(3, 3));

    top = new Panel();
    top.setLayout(new BorderLayout(2, 2));

    numSystemPanel = new Panel();
    numSystemPanel.setLayout(new BorderLayout(2, 2));
    
    


    /* Make the area where the Mantissa and exponent are
       specified. */
    input = new Panel();
    input.setLayout(new GridLayout(1,0,2,2));  
      
    /* Set up the input fields */
    mantSize = new ScrollableNumbers (mantSizeDef, /* initial value */
				      1, /* small increment */
				      0, /* large increment */
				      1, /* lower limit */
				      4, /* upper limit */
				      2,    /* number of chars */
				      /* location of label */
				      ScrollableNumbers.TOP,
				      "Mantissa Size", /* label */
				      /* who to notify when this field
					 changes */
				      this
				     );
				      
    mantSize.setEditable(false);

    minExp = new ScrollableNumbers (minExpDef, /* initial value */
				      1, /* small increment */
				      0, /* large increment */
				      -5, /* lower limit */
				      5, /* upper limit */
				      2,    /* number of chars */
				      /* location of label */
				      ScrollableNumbers.TOP,
				      "Min Exponent", /* label */
				      /* who to notify when this field
					 changes */
				      this
				     );
				      
				      

    minExp.setEditable(false);
    

    maxExp = new ScrollableNumbers (maxExpDef, /* initial value */
				      1, /* small increment */
				      0, /* large increment */
				      -5, /* lower limit */
				      5, /* upper limit */
				      2,    /* number of chars */
				      /* location of label */
				      ScrollableNumbers.TOP,
				      "Max Exponent", /* label */
				      /* who to notify when this field
					 changes */
				      this
				     );


    maxExp.setEditable(false);


    /* Top Label for the mantissa and exp specs */
    Label label;

    input.add(mantSize);
    input.add(minExp);
    input.add(maxExp);

    /* Now create the left panel */

    numSystemPanel.add("North", input);

    eqn = new Eqn(this);
    top.add("Center", numSystemPanel);
    top.add("South", eqn);

    /* The enumeration text areas */
    enum = new Panel();
    enum.setLayout(new BorderLayout(2, 2));

    enumeration = new Panel();
    enumeration.setLayout(new BorderLayout(2, 2));
    
    mantEnum = new TextArea(4, 25);  
    mantEnum.setEditable(false);
    floatEnum = new TextArea(5, 25);
    floatEnum.setEditable(false);
    enumeration.add("North", mantEnum);
    enumeration.add("Center", floatEnum);
    enum.add("Center", enumeration);

    /* **bottom panel** */
    bottom = new Panel();

    bottom.setLayout(new BorderLayout());
   

    plot.resize(500,100);
    plot.setForeground(Color.cyan);
    plot.setBackground(Color.gray);
    bottom.add("Center", plot);

    incrPlot = new Button (">>");
    bottom.add("East", incrPlot);
    decrPlot = new Button ("<<");
    bottom.add("West", decrPlot);

    /*System.out.println("Plot Size is " + plot.size());
    System.out.println("maxint is " + Integer.MAX_VALUE);*/
    statusLine = new Label("This is the Status Line " + 
			   "	                    ");
    statusLinePanel = new Panel();
    statusLinePanel.add(statusLine);

    add("North", top);
    add("Center", enum);
    add("South", bottom);

    windowMenu = new Menu("Window");
    windowMenu.add( executionItem = new MenuItem("Execution"));

    menubar = getMenuBar();
    menubar.add(windowMenu);

    pack();
  }


  public Dimension preferredSize()  {
    return new Dimension(450,550);
  }

  public synchronized void run() {

    /*  public void doCalc(boolean redoPlot_f) {*/
    int mant, exp, maxexp, minexp, expindex, i, j;
    double stepsize, curmant, dnumsteps;
    int numsteps, curexp;

    /*    showStatus("Calculating the mantissa and floating point numbers");*/
    mant = mantSize.getIntText();
    minexp = minExp.getIntText();
    maxexp = maxExp.getIntText();

    if (minexp >  maxexp) {
      minexp = maxexp;
      minExp.setDoubleText(minexp);
    }

    stepsize = 10 / Math.pow (10, mant);
    dnumsteps = Math.pow(10, mant) - Math.pow(10, mant - 1);
    numsteps = new Double(dnumsteps).intValue();

    /*System.out.println(stepsize + " : " + numsteps + " : " + minexp
		       + " : " + maxexp);*/

    /* The lists of mantissas and floating-point numbers get 
       unwieldy when they get large.  Could you instead 
       display it using ranges?  For example, when the mantissa 
       size is 2, the minimum exponent -1, and the maximum 
       exponent 1, do something like

       Mantissas:		Numbers:
       1.0 to 9.9 by 0.1	 .10 to   .99 by  .01
				 1.0 to  9.9  by  .1
         			 10 to   99    by 1.0 
     */
    int base = 10;
    double begin = 1.0;
    double end;
    mantEnum.setText("");
    curmant = 1;
    end = begin + (numsteps-1) * stepsize;

    /* Stop the plotting if it is going on. */
    if (plot.canvasThread != null)
      plot.canvasThread.stop ();
    plot.canvasThread = null;

    /* ERROR-CHECK Also, check the size of the mantSize and the 
     * difference in exponents.  If the combo is too large,
     * show an error on the status line (? and change one)
     */

    /* ERROR-CHECK If the mant-size is greater than 3, change the enumType
     *  to summary (including the button), so that the list doesn't
     *  get too long 
     */
    mantEnum.appendText("Possible Mantissas\n\n");

    mantEnum.appendText("1");
    if (mant > 1) mantEnum.appendText(".");
    for (i = mant; i > 1; i--)
	mantEnum.appendText("0");

    mantEnum.appendText(", " +
			Double.toString(1.0+stepsize) + ", " +
			Double.toString(1.0+2*stepsize) + " ..., " +
			Double.toString(10-3*stepsize) + ", " +
			Double.toString(10-2*stepsize) + ", " +
			Double.toString(10-stepsize));

    
    /*System.out.println("");*/

    floatEnum.setText("");
    double xmin, xmax, floatBegin, floatEnd, floatStepSize;
    counter = 0;
    floatEnum.appendText("Possible Floating Point Numbers\n\n"); 
    /* Now, get all the numbers. */
    for (exp = minexp; exp <= maxexp; exp++) {
      /* We want a comma between all entries, but not at the beginning */
      if (counter != 0)
	floatEnum.appendText(",\n"); 
      
      floatEnum.appendText(truncate(1*Math.pow(10,exp)));

      floatEnum.appendText(
          ", " +
	  truncate((1+stepsize)*Math.pow(10,exp)) +
   	  ", " +
	  truncate((1+2*stepsize)*Math.pow(10,exp)) +
	  " ..., " +
	  truncate((10-3*stepsize)*Math.pow(10,exp)) +
	  ", " +
	  truncate((10-2*stepsize)*Math.pow(10,exp)) +
	  ", " +
	  truncate((10-1*stepsize)*Math.pow(10,exp)));

      counter++;
      appThread.yield();

    }

    if (doPlotInCalc == true) {

	/* This needs to be changed */

	xmin = Math.pow(10,minexp);
	xmax = (10-stepsize)*Math.pow(10,maxexp);


      /* Really need just one call, since all of these need to
	 be done (well, not xmin and xmax, but those two should
	 be together). */

	stateXmin = xmin;
	stateXmax = xmax;

        plot.setXmin(xmin);
        plot.setXmax(xmax);
	plot.setVals(minexp, maxexp, stepsize, mant);

	plot.repaint();
    }

    sharedInfo.mantSize = mant;
    sharedInfo.minExp = minexp;
    sharedInfo.maxExp = maxexp;
    eqn.parseIt();
      
  } /* End of run (doCalc) */



  private String truncate (double x) {
    return(Double.toString(Math.round(x*100000000)/100000000.));
  }


  /* Reset the applet (called from the File menu).  
   * In this applet, it just resets the max and min of the plot
   * to what is should be for the state.
   */
  public void reset () {
      plot.setXmin(stateXmin);
      plot.setXmax(stateXmax);
      plot.repaint();
    }

  public boolean action (Event e, Object arg) {
    int newFn;

    if (super.action(e, arg)) return(true);

    try {
	if ( e.target == mantSize || e.target == minExp
	    || e.target == maxExp) {
	  doPlotInCalc = true;
	  if (appThread != null)
	    appThread.stop();
	  appThread = new Thread(this);
	  appThread.start();
	  plot.repaint();
	  return true;
	}
	/*if ((e.target instanceof Button) || 
	      (e.target instanceof TextField)) {
	  		 && (e.id == Event.ACTION_EVENT)) {
	  if (e.target == recalc) { 
	  doPlotInCalc = true;
	  if (appThread != null)
	    appThread.stop();
	  appThread = new Thread(this);
	  appThread.start();
	  plot.repaint();
	  return true;
	}*/
	else if (e.target == incrPlot) {
	  double xmax;

	  plot.setXmin(0);
	  plot.setXmax(plot.getXmax() * 10);
	  plot.repaint();
	  return true;
	}
	else if (e.target == decrPlot) {
	  double xmax;

	  plot.setXmin(0);
	  plot.setXmax(plot.getXmax() / 10);
	  plot.repaint();
	  return true;
	}
      else if (e.target.equals( executionItem))
	eqn.rdsFrame.show();
	
      }
    catch (NumberFormatException e1) {
      statusLine.setText("Domain must be a number");
      /*System.out.println(e1);*/
    }
  return false;
 } /* end action() */ 

} /* end of FP class */


class MyCanvas extends Panel implements Runnable
{

  /* May need to use threads to be sure canvas is repainted after init */

  int inset = 20;
  double ymax, ymin, xmax, xmin;

  int radius = 3;               // Size of the circles
  /* These are used in the axis */
  double xaxisPosy, yaxisPosx;
  int tickHalfLen = 2;
  PlotCoord pc;
  /* for rubberbanding when selection new area */
  double tempMouseDown;
  int xfrom;
  int oldTox, oldToy;
  int yfrom, yto;
  int xprev; 
  boolean prevSet;
  Thread canvasThread = null;

    double maxExp, minExp, stepSize;
    int mantLength;

  MyCanvas() {
    super();
    xmin = -100;
    xmax = 100;
    ymin = -2;
    ymax = 2;
  }


  public Dimension preferredSize () {
    FontMetrics fm = getFontMetrics(getFont());
    return(new Dimension(500, 100));
  }


  public void setVals (double newMinExp, double newMaxExp,
		       double newStepSize, int newMantLength) {
    minExp = newMinExp;
    maxExp = newMaxExp;
    stepSize = newStepSize;
    mantLength = newMantLength;
  }

  public void setXmin(double newXmin) {
    xmin = newXmin;
  }

  public double getXmin() {
    return(xmin);
  }

  public void setXmax(double newXmax) {
    xmax = newXmax;
  }

  public double getXmax() {
    return(xmax);
  }

  /*  PAINT
   */ 
  /** Draws a box around this panel, plus other stuff. */
  public void paint(Graphics g) {

    /* if it's already trying to draw, get rid of the old drawing */
    if (canvasThread != null) {
      canvasThread.stop();
    }
    canvasThread = null;
    canvasThread = new Thread(this);
    canvasThread.start();
  } /* end paint */


  public void run() {
    int x, y;
    Dimension d = size();
    /* These need to be made into floats. */
    double xcur, ycur = 1;
    int xcoord, ycoord;
    Graphics g;

    System.out.println("");
/*    try {
      Thread.sleep(100);
    } catch (InterruptedException e) {}*/
    System.out.println("");
    g = getGraphics();
    System.out.println("");

    /*System.out.println("got a graphic");*/
    /* Draw a rectangle around the Canvas */
    g.drawRect(0,0, d.width - 1, d.height - 1);

    /* Set up the transformation between cartesian coordinates
       and screen coordinates. */
    pc = new PlotCoord (xmax, ymax, xmin, ymin, d.width, d.height, inset);

    /* If there is only one value for y, set up a little bit of room
       in the graph */
    if (ymax == ymin) {
      ymax = ymax + 2;
      ymin = ymin - 2;
    }

    drawAxis(g);

    /* In this case, we know we are always plotting at one */
    ycoord = pc.ycoord(1);

    int i, j;

    int xcoordPrev = -1;


    for (double exp = minExp; exp <= maxExp; exp++) {

	double power = Math.pow(10, exp);


	// Find the largest and smallest numbers to be displayed

	double smallest = power;
	double largest = (10 - stepSize) * power;

	largest = Math.min(largest, xmax);
	smallest = Math.max(smallest, xmin);

	largest = Math.floor(largest / power * Math.pow(10,mantLength-1));
	smallest = Math.ceil(smallest / power * Math.pow(10,mantLength-1));
	int numLines = (int) Math.round((largest - smallest) + 1);

	largest = largest / Math.pow(10,mantLength-1) * power;
	smallest = smallest / Math.pow(10,mantLength-1)*power;

	if (largest < xmin) continue;
	if (smallest > xmax) return;
	if (largest < smallest) continue;

	// See if the entire decade can be drawn as a rectangle

	int lowCoord = pc.xcoord(smallest);
	int highCoord = pc.xcoord(largest);

	if (highCoord - lowCoord <= numLines - 1)
	    g.fillRect(lowCoord, ycoord-2, highCoord-lowCoord+1, 5);

	else {
	    double step = (highCoord - lowCoord) / (numLines - 1.0);
	    double coord = lowCoord;
	    for (int k = 0; k < numLines; k++) {
		g.drawLine((int) Math.round(coord), ycoord-2,
			   (int) Math.round(coord), ycoord+2);
		coord = coord + step;
	    }
	}

    }

  }



  /* This is for when you want to zero in on part of the graph */
  public boolean mouseDown(Event e, int xpos, int ypos) {

    yfrom = pc.ycoord(-2.5);
    yto   = pc.ycoord( 2.5);

    xfrom = xpos;
    tempMouseDown = pc.coordToX(xpos);
    if (tempMouseDown < 0) 
      tempMouseDown = 0;
    if (tempMouseDown > xmax)
      tempMouseDown = xmax;
    return true;
  }

  public boolean mouseDrag(Event e, int xpos, int ypos) {
    Graphics g = getGraphics();
    int xto;

    xto = xpos;

    g.setXORMode(Color.pink);
    if (prevSet) {
      drawBox(xfrom, xprev, yfrom, yto, g);
    }
    drawBox(xfrom, xto, yfrom, yto, g);
    getToolkit().sync();
    xprev = xto;
    prevSet = true;
    g.setPaintMode();
    return true;
  }


  void drawBox(int xfrom, int xto, int yfrom, int yto, Graphics g) {
    g.drawLine(xfrom, yfrom, xfrom, yto);
    g.drawLine(xfrom, yto, xto, yto);
    g.drawLine(xto, yto, xto, yfrom);
    g.drawLine(xto, yfrom, xfrom, yfrom);
  }


  public boolean mouseUp(Event e, int xpos, int ypos) {
    Graphics g = getGraphics();
    double tempMouseUp;

    prevSet = false;
    tempMouseUp = pc.coordToX(xpos);
    if (tempMouseUp < 0) 
      tempMouseUp = 0;
    if (tempMouseUp > xmax)
      tempMouseUp = xmax;
    
    /* If they just click in one spot, don't change anything */
    if (tempMouseDown == tempMouseUp)
      return true;

    if (tempMouseDown < tempMouseUp) {
      xmin = tempMouseDown;
      xmax = tempMouseUp;
    }
    else {
      xmin = tempMouseUp;
      xmax = tempMouseDown;
    }
    
    repaint();
    return true;
  }

  /* DRAW AXIS
   */
  void drawAxis(Graphics g)
  {

    double xaxisStart = xmax, xaxisEnd = xmin, xaxisPosy;
    double yaxisStart = ymax, yaxisEnd = ymin, yaxisPosx;

    /* Axis rules:
       If 0 is on the map, print out the axis there.
       If 0 is not on the map, print out the axis on the 
       minimum. */

    if (xmin < 0  && xmax > 0) 
      yaxisPosx = 0;
    else
      yaxisPosx = xmin;

    if (ymin < 0  && ymax > 0) 
      xaxisPosy = 0;
    else
      xaxisPosy = ymin ;


    /* Draw x axis. */
    g.drawLine(pc.xcoord(xaxisStart), pc.ycoord(xaxisPosy), 
	       pc.xcoord(xaxisEnd), pc.ycoord(xaxisPosy));
    /* Draw y axis. */
/*    g.drawLine(pc.xcoord(yaxisPosx), pc.ycoord(yaxisStart),
	       pc.xcoord(yaxisPosx), pc.ycoord(yaxisEnd));*/

    Font f = new Font ("Dialog", Font.PLAIN, 10);
    g.setFont(f);

    drawAllTicks(g);

  }


  void drawAllTicks(Graphics g) {
    double mult, multPart, dif, curTick, logNum = 0;
    int i = 0;
    double foo = 0;
    double lwr;

    /* draw a tick at 0 if it is on the graph */
    if (xmin <= 0  && xmax >= 0) {
      drawXTick(0, g);
    }

    /* draw x ticks */
    dif = xmax - xmin;
    
    try {
      logNum = Math.log(dif)/Math.log(10);
    } catch (ArithmeticException e) {}

    logNum = Math.floor(logNum);

    /*System.out.println("xmax is " + xmax + ", xmin is " + xmin + 
                ", i is " + i + ", logNum is " + logNum + ", dif is " + dif);*/

    if (dif/Math.pow(10,logNum) < 3) {
      foo = 2;
    }

    /* System.out.println("xmax is " + xmax + ", xmin is " + xmin + 
               ", i is " + i + ", logNum is " + logNum + ", dif is " + dif);*/

    if (foo == 0)
      mult = Math.pow(10, logNum);
    else
      mult = 2 * Math.pow(10, logNum-1);
    /*System.out.println("Mult is " + mult);*/

    lwr = xmin / mult;
    lwr = Math.ceil(lwr);
    lwr = lwr * mult;

    for ( curTick = lwr; curTick <= xmax; 
	                        curTick += mult) {
       drawXTick(curTick, g);
     }
  }

  void drawYTick(double y, Graphics g) {
    int xcoord, ycoord;
    int offset;

    ycoord = pc.ycoord(y);
    xcoord = pc.xcoord(yaxisPosx);
    
    if (y == 0)
      offset = 10;
    else
      offset = 5;

    g.drawLine(xcoord + tickHalfLen, 
	       ycoord , xcoord - tickHalfLen, ycoord);
    g.drawString(new Float(y).toString(),
		 xcoord + 5, ycoord + offset);

/*    g.drawLine(xcoord + tickHalfLen, ycoord , xcoord - tickHalfLen, ycoord);
    g.drawString(new Float(y).toString(),
		 xcoord + 5, ycoord + 10);*/
  }

  void drawXTick(double x, Graphics g) {
    int xcoord, ycoord;

    xcoord = pc.xcoord(x);
    ycoord = pc.ycoord(xaxisPosy);

    g.drawLine(xcoord, ycoord + tickHalfLen, 
	       xcoord , ycoord - tickHalfLen);
    if (x == 0)
       g.drawString(new Float(x).toString(),
		   xcoord, ycoord + 15);
    else
       g.drawString(new Float(x).toString(),
		   xcoord - 5, ycoord + 15);
      

  }

} /* End MyPlot */



/* CLASS        PLOT COORD
 *
 * 
 */
class PlotCoord {
  
  double xstretch = 1;
  double ystretch = 1;
  double xshift = 0;
  double yshift = 0;
  double ywholeScreen;       
  int inset;             /* How far from the edge the display is */

  PlotCoord(double xmax, double ymax, double xmin, double ymin, 
	    int width, int height, int insetin ) {

    inset = insetin;

    xstretch = (width - (2 * inset))/ (xmax - xmin);
    ystretch = (height - (2 * inset)) / (ymax - ymin);

    xshift = -xmin;
    yshift = -ymin;

    ywholeScreen = ymax - ymin;

  }

  /* Return the X and Y coordinates on the canvas. */
  int xcoord(double x) {
    double xcoordFloat;
    
    xcoordFloat = (x + xshift) * xstretch + inset;
    return(new Long(Math.round(xcoordFloat)).intValue());
  }

  int ycoord(double y) {
    double ycoordFloat;
    
    ycoordFloat = (ywholeScreen - (y + yshift)) * ystretch + inset;
    return(new Long(Math.round(ycoordFloat)).intValue());
  }
       
  /* Return the X and Y real coords on canvas. */
  double coordToX(int xcoord) {
    double x;

    x = (xcoord - inset) / xstretch - xshift;
    return(x);
  }

  double coordToY(int ycoord) {
    double y;

    y = ((ycoord - inset) / ystretch) + ywholeScreen - yshift;
    return(y);
  }
       

}

class FPSharedInfo {
  int mantSize,
      minExp,
      maxExp;
}
